#include "vqmefound.h"
#include "stralloc.h"
#include "constmap.h"
#include "open.h"
#include "case.h"
#include "byte.h"
#include "wait.h"
#include "slurpclose.h"
#include <sys/types.h>
#include <sys/stat.h>
#include "exit.h"
#include "fork.h"
#include <unistd.h>
#include "cdb.h"
#include "error.h"
#include "control.h"
#include "fd.h"
#include "qlx.h"

static int file_exists(const char *fn)
{
  struct stat st;
  if(fn == 0) return -1;
  if(stat(fn,&st) == -1){
    if(errno == error_noent) return 0;
  }else return 2;

  return 1;
}

int goodlocalscheckok = 0;
static stralloc goodlocalscheck = {0};
static stralloc locals = {0};
static struct constmap maplocals;
int vdomsok = 0;
static stralloc vdoms = {0};
static struct constmap mapvdoms;

int blcok = 0;
static stralloc blc = {0};
static struct constmap mapblc;

static int vqmefound_init_send()
{
  if(file_exists("control/goodlocalscheck") != 2){
    goodlocalscheckok = 0;
    return 1;
  }
  goodlocalscheckok = control_readfile(&goodlocalscheck,"control/goodlocalscheck",0);
  if (goodlocalscheckok == -1) return 0;
  if (!goodlocalscheckok) return 0;
  if (control_readfile(&locals,"control/locals",1) != 1) return 0;
  if (!stralloc_cat(&locals,&goodlocalscheck)) return 0;
  if (!constmap_init(&maplocals,locals.s,locals.len,0)) return 0;
  vdomsok = control_readfile(&vdoms,"control/virtualdomains",0);
  if(vdomsok == -1) return 0;
  if(vdomsok)
    if (!constmap_init(&mapvdoms,vdoms.s,vdoms.len,1)) return 0;

  blcok = control_readfile(&blc,"control/badlocalscheck",0);
  if(blcok == -1) return 0;
  if(blcok)
    if (!constmap_init(&mapblc,blc.s,blc.len,0)) return 0;

 return 1;
}

static int vqmefound_init_lspawn()
{
  if (!goodlocalscheckok) return 1;
  return 1;
}

static int vqmefound_init_local()
{
  if (!goodlocalscheckok) return 1;
  return 1;
}

/* error = -1 */
int vqmefound_init()
{
  if(!vqmefound_init_send()) return -1;
  if(!vqmefound_init_lspawn()) return -1;
  if(!vqmefound_init_local()) return -1;
  return 0;
}

static stralloc localuser = {0};
static int vqmefound_mailaddr_to_localuser(char *buf,int len)
{
  int at;
  int i;
  char *x;

  at = byte_rchr(buf,len,'@');
  if(at == len){
    if (!stralloc_copyb(&localuser,buf,len)) return -1;
    if (!stralloc_0(&localuser)) return -1;
    return 1;
  }
  if(at < 1 || at > len) return 0;

  if(blcok)
    if (constmap(&mapblc,buf + at + 1,len - at - 1)) return 2;

  if (constmap(&maplocals,buf + at + 1,len - at - 1)) {
    if (!stralloc_copyb(&localuser,buf,at)) return -1;
    if (!stralloc_0(&localuser)) return -1;
    return 1;
  }

  if(vdomsok){
    for (i = 0;i <= len;++i)
      if (!i || (i == at + 1) || (i == len) || ((i > at) && (buf[i] == '.')))
        if (x = constmap(&mapvdoms,buf + i,len - i)) {
          if (!*x) break;
          if (!stralloc_copys(&localuser,x)) return -1;
          if (!stralloc_cats(&localuser,"-")) return -1;
          if (!stralloc_catb(&localuser,buf,at)) return -1;
          if (!stralloc_0(&localuser)) return -1;
          return 1;
        }
  }

  return 0;
}

static stralloc lower = {0};
static stralloc nughde = {0};
static stralloc wildchars = {0};

static int vqmefound_localuser_to_nughde_cdb(int fd)
{
   int r;
   int flagwild;
   uint32 dlen;
   unsigned int i;

   r = cdb_seek(fd,"",0,&dlen);
   if (r != 1) return -1;
   if (!stralloc_ready(&wildchars,(unsigned int) dlen)) return -1;
   wildchars.len = dlen;
   if (cdb_bread(fd,wildchars.s,wildchars.len) == -1) return -1;

   i = lower.len;
   flagwild = 0;

   do
    {
     /* i > 0 */
     if (!flagwild || (i == 1) || (byte_chr(wildchars.s,wildchars.len,lower.s[i - 1]) < wildchars.len))
      {
       r = cdb_seek(fd,lower.s,i,&dlen);
       if (r == -1) return -1;
       if (r == 1)
        {
         if (!stralloc_ready(&nughde,(unsigned int) dlen)) return -1;
         nughde.len = dlen;
         if (cdb_bread(fd,nughde.s,nughde.len) == -1) return -1;
         if (flagwild){
           if(localuser.len <= i-1) return -1;
           if (!stralloc_cats(&nughde,localuser.s)) return -1;
         }
         if (!stralloc_0(&nughde)) return -1;
         return 1;
        }
      }
     --i;
     flagwild = 1;
    }
   while (i);

   return 0;
}

static stralloc qme = {0};

static int vqmefound_file_exists()
{
  if (!stralloc_0(&qme)) return -1;
  return file_exists(qme.s);
}

static int vqmefound_localuser_to_nughde()
{
 int pi[2];
 char *(args[3]);
 int gpwpid;
 int gpwstat;
 int fd;

 if (!stralloc_copys(&lower,"!")) return -1;
 if (!stralloc_cats(&lower,localuser.s)) return -1;
 if (!stralloc_0(&lower)) return -1;
 case_lowerb(lower.s,lower.len);

 if (!stralloc_copys(&nughde,"")) return -1;
 fd = open_read("users/cdb");
 if (fd == -1)
   if (errno != error_noent)
     return -1;

 if(fd != -1){
   switch(vqmefound_localuser_to_nughde_cdb(fd)){
     case -1: close(fd); return -1;
     case 1: close(fd); return 1;
   }
   close(fd);
 }else{
   if (!stralloc_copys(&qme,"users/cdb")) return -1;
   if (vqmefound_file_exists()) return -1;
 }

 if (pipe(pi) == -1) return -1;
 args[0] = "bin/qmail-getpw";
 args[1] = localuser.s;
 args[2] = 0;
 switch(gpwpid = vfork())
  {
   case -1:
     return -1;
   case 0:
     close(pi[0]);
     if (fd_move(1,pi[1]) == -1) _exit(QLX_SYS);
     execv(*args,args);
     _exit(QLX_EXECPW);
  }
  close(pi[1]);

  if (slurpclose(pi[0],&nughde,128) == -1){
    wait_pid(&gpwstat,gpwpid);
    return -1;
  }

  if (wait_pid(&gpwstat,gpwpid) != -1){
    if (wait_crashed(gpwstat)) return -1;
    if (wait_exitcode(gpwstat) != 0) return -1;
  }
  return 1;
}

static stralloc localhome = {0};
static stralloc localdash = {0};
static stralloc localext = {0};

static int vqmefound_localuser_to_hde()
{
  char *x;
  char *tp;
  unsigned int xlen;
  int n;
  int i;

  switch(vqmefound_localuser_to_nughde()){
    case -1: return -1;
    case 0: return 0;
  }
  x = nughde.s;
  xlen = nughde.len;

  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* user */
  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* uid */
  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* gid */

  tp = x;
  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* homedir */
  if (!stralloc_copys(&localhome,tp)) return -1;

  tp = x;
  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* dash */
  if (!stralloc_copys(&localdash,tp)) return -1;
  tp = x;
  n = byte_chr(x,xlen,0); if (n++ == xlen) return -1; x += n; xlen -= n; /* ext */
  if (!stralloc_copys(&localext,tp)) return -1;

  case_lowerb(localext.s,localext.len);
  for (i = 0;i < localext.len;++i)
    if (localext.s[i] == '.')
      localext.s[i] = ':';
  return 1;
}

static int vqmefound_hde_to_found()
{
  int i;
  if (localdash.len == 0) return 1;
  if (localdash.len < 1 || localdash.s == 0) return -1;
  if (*(localdash.s) == '\0') return 1;
  if (!stralloc_copy(&qme,&localhome)) return -1;
  if (!stralloc_cats(&qme,"/.qmail")) return -1;
  if (!stralloc_cat(&qme,&localdash)) return -1;
  if (!stralloc_cat(&qme,&localext)) return -1;
  switch(vqmefound_file_exists()){
    case -1: return -1;
    case 0: break;
    default : return 1;
  }

  for (i = localext.len;i >= 0;--i)
    if (!i || (localext.s[i - 1] == '-')) {
      if (!stralloc_copy(&qme,&localhome)) return -1;
      if (!stralloc_cats(&qme,"/.qmail")) return -1;
      if (!stralloc_cat(&qme,&localdash)) return -1;
      if (!stralloc_catb(&qme,localext.s,i)) return -1;
      if (!stralloc_cats(&qme,"default")) return -1;
      switch(vqmefound_file_exists()){
        case -1: return -1;
        case 0: break;
        default: return 1;
      }
    }

  return 0;
}

/* error -1 , mailaddr = buf, buf == user unknown,0 else 1 */
int vqmefound(buf,len)
char *buf;
int len;
{
  if(len == 0) return 0;
  if(buf == 0 || len < 1) return -1;
  if (!goodlocalscheckok) return -2;
  switch(vqmefound_mailaddr_to_localuser(buf,len)){
    case 1: break;
    case 0: return -2;
    case 2: return 0;
    default: return -1;
  }
  switch(vqmefound_localuser_to_hde()){
    case 1: break;
    case 0: return 0;
    default: return -1;
  }
  return vqmefound_hde_to_found();
}
